import aiohttp_jinja2
from aiohttp import web

from logging import getLogger
_log = getLogger(__name__)


class Clients():
    def __init__(self, app, db=None, security=None):
        self._db_conn = db.conn
        self._security = security
        app.router.add_get('/clients', self._index, name='clients')
        app.router.add_get('/clients/new', self._new, name='clients_new')
        app.router.add_post('/clients/new', self._new_apply,)
        app.router.add_get('/clients/{item_id}',
                           self._view, name='clients_view')
        app.router.add_get('/clients/{item_id}/edit',
                           self._edit, name='clients_edit')
        app.router.add_post('/clients/{item_id}/edit', self._edit_apply)
        app.router.add_get('/clients/{item_id}/delete',
                           self._delete, name='clients_delete')

    @ aiohttp_jinja2.template("clients/index.html")
    async def _index(self, request: web.Request):
        await self._security.raise_perm(request, perm='superuser')
        cursor = self._db_conn.cursor()
        cursor.execute('SELECT * FROM clients')
        items = cursor.fetchall()
        return {"items": items}

    @ aiohttp_jinja2.template("clients/new.html")
    async def _new(self, request: web.Request):
        await self._security.raise_perm(request, perm='superuser')
        return {}

    async def _new_apply(self, request: web.Request):
        await self._security.raise_perm(request, perm='superuser')
        item = await request.post()
        cursor = self._db_conn.cursor()
        try:
            cursor.execute(
                ''' INSERT INTO clients
                    (cloud_identifier, is_disabled,
                    public_user_key, ssh_server_listen_port,
                    ssh_server_hostname, ssh_server_port)
                    VALUES (%s, %s, %s, %s, %s, %s) ''',
                (item['cloud_identifier'], item['is_disabled'],
                 item['public_user_key'], item['ssh_server_listen_port'],
                 item['ssh_server_hostname'], item['ssh_server_port']))
            self._db_conn.commit()
        except Exception as e:
            self._db_conn.rollback()
            _log.debug(f'clients add: {e}')
            raise web.HTTPSeeOther(request.app.router['clients_new'].url_for())
        item_id = str(cursor.lastrowid)
        raise web.HTTPSeeOther(
            request.app.router['clients_view'].url_for(item_id=item_id))

    @ aiohttp_jinja2.template("clients/view.html")
    async def _view(self, request: web.Request):
        await self._security.raise_perm(request, perm='superuser')
        item_id = request.match_info["item_id"]
        cursor = self._db_conn.cursor()
        cursor.execute('SELECT * FROM clients WHERE id = %s', (item_id,))
        item = cursor.fetchone()
        if item is None:
            raise web.HTTPSeeOther(request.app.router['clients'].url_for())
        return {"item": item}

    @ aiohttp_jinja2.template("clients/edit.html")
    async def _edit(self, request: web.Request):
        await self._security.raise_perm(request, perm='superuser')
        item_id = request.match_info["item_id"]
        cursor = self._db_conn.cursor()
        cursor.execute('SELECT * FROM clients WHERE id = %s', (item_id,))
        item = cursor.fetchone()
        if item is None:
            raise web.HTTPSeeOther(request.app.router['clients'].url_for())
        return {"item": item}

    async def _edit_apply(self, request: web.Request):
        await self._security.raise_perm(request, perm='superuser')
        item_id = request.match_info["item_id"]
        item = await request.post()
        cursor = self._db_conn.cursor()
        try:
            cursor.execute(
                '''UPDATE clients
                        SET cloud_identifier = %s, is_disabled = %s,
                        public_user_key = %s, ssh_server_listen_port = %s,
                        ssh_server_hostname = %s, ssh_server_port = %s,
                        WHERE id = %s ''',
                (item['cloud_identifier'], item['is_disabled'],
                 item['public_user_key'], item['ssh_server_listen_port'],
                 item['ssh_server_hostname'], item['ssh_server_port'],
                 item_id))
            cursor.commit()
        except Exception as e:
            _log.debug(f'clients: edit fail: {e}')
            raise web.HTTPSeeOther(
                request.app.router['clients_edit'].url_for(item_id=item_id))
        if cursor.rowcount < 1:
            _log.debug('clients: edit fail:')
            raise web.HTTPSeeOther(
                request.app.router['clients_edit'].url_for(item_id=item_id))
        raise web.HTTPSeeOther(
            request.app.router['clients_view'].url_for(item_id=item_id))

    async def _delete(self, request: web.Request):
        await self._security.raise_perm(request, perm='superuser')
        item_id = request.match_info["item_id"]
        if item_id is None or len(item_id) < 1:
            raise web.HTTPSeeOther(request.app.router['clients'].url_for())
        cursor = self._db_conn.cursor()
        try:
            cursor.execute(
                'DELETE FROM clients WHERE id = %s',
                (item_id,))
            self._db_conn.commit()
        except Exception as e:
            self._db_conn.rollback()
            _log.debug(f'clients delete fail: {e}')
            raise web.HTTPSeeOther(
                request.app.router['clients_delete'].url_for())
        if cursor.rowcount < 1:
            _log.debug('clients delete fail')
            raise web.HTTPSeeOther(
                request.app.router['clients_delete'].url_for())
        raise web.HTTPSeeOther(request.app.router['clients'].url_for())



import aiohttp_jinja2


class Home():
    def __init__(self, app, db=None, security=None, ssh=None):
        self._db_conn = db.conn
        self._ssh = ssh

        self._security = security
        app.router.add_get('/', self.index, name='home')

    @aiohttp_jinja2.template('home/index.html')
    async def index(self, request):
        await self._security.raise_perm(request, perm="superuser")
        clients = self._ssh.get_clients()
        return {'clients':  clients}



import psycopg2
from psycopg2.extras import DictCursor
import asyncio
from pathlib import Path
from yoyo import read_migrations
from yoyo import get_backend

from logging import getLogger
_log = getLogger(__name__)

_instance = None


def get_db():
    global _instance
    return _instance


class Db:
    def __init__(self, options):
        global _instance
        if _instance is None:
            _instance = self

        self.conn = None

        self._do_migrations(options)

        try:
            self.conn = psycopg2.connect(
                dbname=options.db_name,
                user=options.db_user,
                cursor_factory=DictCursor)
        except psycopg2.Error as e:
            _log.error(f'Unable to connect to database: {e}')
            if self.conn is not None:
                self.conn.close()
            exit(1)

    def _do_migrations(self, options):
        try:
            backend = get_backend(
                f'postgresql://{options.db_user}@/{options.db_name}'
            )
        except psycopg2.Error as e:
            _log.error(f'Unable to open database: {e}')
            exit(1)

        migrations = read_migrations(options.db_migrations_dir)
        with backend.lock():
            backend.apply_migrations(backend.to_apply(migrations))

    def __del__(self):
        try:
            if self.conn is not None:
                self.conn.close()
        except psycopg2.Error as e:
            _log.error(f'database.py: Unable to close Databse: {e}')



import aiohttp_jinja2
from aiohttp import web

from logging import getLogger
_log = getLogger(__name__)


class Users():
    def __init__(self, app, db=None, security=None):
        self._db_conn = db.conn
        self._security = security
        app.router.add_get('/users', self._index, name='users')
        app.router.add_get('/users/new', self._new, name='users_new')
        app.router.add_post('/users/new', self._new_apply)
        app.router.add_get('/users/{item_id}',
                           self._view, name='users_view')
        app.router.add_get('/users/{item_id}/edit',
                           self._edit, name='users_edit')
        app.router.add_post('/users/{item_id}/edit', self._edit_apply)
        app.router.add_get('/users/{item_id}/delete',
                           self._delete, name='users_delete')

    @ aiohttp_jinja2.template("users/index.html")
    async def _index(self, request: web.Request):
        await self._security.raise_perm(request, perm='superuser')
        cursor = self._db_conn.cursor()
        cursor.execute('SELECT * FROM users')
        items = cursor.fetchall()
        return {"items": items}

    @ aiohttp_jinja2.template("users/new.html")
    async def _new(self, request: web.Request):
        await self._security.raise_perm(request, perm='superuser')
        return {}

    async def _new_apply(self, request: web.Request):
        await self._security.raise_perm(request, perm='superuser')
        item = await request.post()
        cursor = self._db_conn.cursor()
        try:
            cursor.execute(
                ''' INSERT INTO users
                    (identity, password,
                    full_name, email,
                    permission_id, is_superuser,
                    is_disabled, remember_me)
                    VALUES (%s, %s, %s, %s, %s, %s, %s, %s) ''',
                (item['identity'], item['password'],
                 item['full_name'], item['email'],
                 item['permission_id'], item['is_superuser'],
                 item['is_disabled'], item['remember_me']))
            self._db_conn.commit()
        except Exception as e:
            self._db_conn.rollback()
            _log.debug(f'users add: {e}')
            raise web.HTTPSeeOther(request.app.router['users_new'].url_for())
        item_id = str(cursor.lastrowid)
        raise web.HTTPSeeOther(
            request.app.router['users_view'].url_for(item_id=item_id))

    @ aiohttp_jinja2.template("users/view.html")
    async def _view(self, request: web.Request):
        await self._security.raise_perm(request, perm='superuser')
        item_id = request.match_info["item_id"]
        cursor = self._db_conn.cursor()
        cursor.execute('SELECT * FROM users WHERE id=%s', (item_id,))
        item = cursor.fetchone()
        if item is None:
            raise web.HTTPSeeOther(request.app.router['users'].url_for())
        return {"item": item}

    @ aiohttp_jinja2.template("users/edit.html")
    async def _edit(self, request: web.Request):
        await self._security.raise_perm(request, perm='superuser')
        item_id = request.match_info["item_id"]
        cursor = self._db_conn.cursor()
        cursor.execute('SELECT * FROM users WHERE id=%s', (item_id,))
        item = cursor.fetchone()
        if item is None:
            raise web.HTTPSeeOther(request.app.router['users'].url_for())
        return {"item": item}

    async def _edit_apply(self, request: web.Request):
        await self._security.raise_perm(request, perm='superuser')
        item_id = request.match_info["item_id"]
        item = await request.post()
        cursor = self._db_conn.cursor()
        try:
            cursor.execute(
                '''UPDATE users
                        SET identity = %s, password = %s,
                        full_name = %s, email = %s,
                        permission_id = %s, is_superuser = %s,
                        is_disabled = %s, remember_me = %s
                        WHERE id = %s ''',
                (item['identity'], item['password'],
                 item['full_name'], item['email'],
                 item['permission_id'], item['is_superuser'],
                 item['is_disabled'], item['remember_me'], item_id))
            cursor.commit()
        except Exception as e:
            _log.debug(f'users: edit fail: {e}')
            raise web.HTTPSeeOther(
                request.app.router['users_edit'].url_for(item_id=item_id))
        if cursor.rowcount < 1:
            _log.debug('users: edit fail:')
            raise web.HTTPSeeOther(
                request.app.router['users_edit'].url_for(item_id=item_id))
        raise web.HTTPSeeOther(
            request.app.router['users_view'].url_for(item_id=item_id))

    async def _delete(self, request: web.Request):
        await self._security.raise_perm(request, perm='superuser')
        item_id = request.match_info["item_id"]
        if item_id is None or len(item_id) < 1:
            raise web.HTTPSeeOther(request.app.router['users'].url_for())
        cursor = self._db_conn.cursor()
        try:
            cursor.execute(
                'DELETE FROM users WHERE id = %s',
                (item_id,))
            self._db_conn.commit()
        except Exception as e:
            self._db_conn.rollback()
            _log.debug(f'users delete fail: {e}')
            raise web.HTTPSeeOther(
                request.app.router['users_delete'].url_for())
        if cursor.rowcount < 1:
            _log.debug('users delete fail')
            raise web.HTTPSeeOther(
                request.app.router['users_delete'].url_for())
        raise web.HTTPSeeOther(request.app.router['users'].url_for())



import aiohttp_jinja2
from aiohttp import web

import asyncio
import sys
import json


class System():
    def __init__(self, app, db=None, security=None, options=None):
        self._db_conn = db.conn
        self._security = security
        self._options = options
        app.router.add_get('/system', self._index, name='system')
        app.router.add_get("/system/check_update",
                           self._check_update, name='system_check_update')
        app.router.add_get("/system/do_update",
                           self._do_update, name='system_do_update')
        app.router.add_get("/system/restart", self._restart,
                           name='system_restart')

    @aiohttp_jinja2.template('system/index.html')
    async def _index(self, request):
        await self._security.raise_perm(request, perm='superuser')
        text = ''
        return {"text": text}

    @aiohttp_jinja2.template('system/index.html')
    async def _check_update(self, request):
        latest_version = "No Upate"
        await self._security.raise_perm(request, perm='superuser')
        proc = await asyncio.create_subprocess_exec(
            sys.executable, '-m', 'pip', 'list', "-o", "--format=json",
            "--disable-pip-version-check",
            "--no-color",
            "--no-python-version-warning",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE)

        stdout, stderr = await proc.communicate()
        data = json.loads(stdout)

        for item in data:
            if item['name'] == 'riegocloud':
                latest_version = item['latest_version']
                break
        return {'text':  latest_version}

    @aiohttp_jinja2.template('system/index.html')
    async def _do_update(self, request):
        await self._security.raise_perm(request, perm='superuser')
        proc = await asyncio.create_subprocess_exec(
            sys.executable, '-m', 'pip', 'install', "riegocloud", "--upgrade",
            "--disable-pip-version-check",
            "--no-color",
            "--no-python-version-warning",
            "-q", "-q", "-q")
        await proc.wait()
        return {'text': "Restart erforderlich"}

    @aiohttp_jinja2.template('system/index.html')
    async def _restart(self, request):
        await self._security.raise_perm(request, perm='superuser')
        # TODO shedule exit for a few seconds and return a redirect
        asyncio.get_event_loop().call_later(1, exit, 0)
        raise web.HTTPSeeOther(request.app.router['system'].url_for())

    async def check_installed(self):
        version = "0.0.0"
        proc = await asyncio.create_subprocess_exec(
            sys.executable, '-m', 'pip', 'list', "--format=json",
            "--disable-pip-version-check",
            "--no-color",
            "--no-python-version-warning",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE)

        stdout, stderr = await proc.communicate()
        data = json.loads(stdout)
        for item in data:
            if item['name'] == 'riegocloud':
                version = item['version']
                break
        return version



from aiohttp import web
import asyncio
from jinja2 import FileSystemLoader, Environment, TemplateNotFound
from pathlib import Path

from logging import getLogger
_log = getLogger(__name__)


class Api():
    def __init__(self, app, db=None, security=None, options=None):
        self._db_conn = db.conn
        self._security = security
        self._options = options
        app.router.add_post(options.http_server_endpoint,
                            self._api_post, name='api')

    async def _api_post(self, request):
        data = await request.json()
        cloud_identifier = data.get('cloud_identifier', '')
        public_user_key = data.get('public_user_key', '')
        if not len(cloud_identifier) or not len(public_user_key):
            await asyncio.sleep(5)
            raise web.HTTPBadRequest

        cursor = self._db_conn.cursor()
        cursor.execute("""SELECT MAX(ssh_server_listen_port) AS max_port
                        FROM clients""")
        max_port = cursor.fetchone()
        max_port = max_port['max_port']
        if max_port is None or max_port == 0:
            ssh_server_listen_port = 33333
        else:
            ssh_server_listen_port = max_port + 1
        try:
            cursor.execute(
                """INSERT INTO clients
                        (cloud_identifier, public_user_key,
                        ssh_server_listen_port,
                        ssh_server_hostname, ssh_server_port)
                        VALUES (%s,%s,%s,%s,%s)""",
                (cloud_identifier, public_user_key,
                 ssh_server_listen_port,
                 self._options.ssh_server_hostname,
                 self._options.ssh_server_port))
            self._db_conn.commit()
        except Exception:
            self._db_conn.rollback()
            cursor.execute('''UPDATE clients SET
                        public_user_key = %s,
                        ssh_server_listen_port = %s,
                        ssh_server_hostname = %s,
                        ssh_server_port = %s
                        WHERE cloud_identifier = %s ''',
                           (public_user_key,
                            ssh_server_listen_port,
                            self._options.ssh_server_hostname,
                            self._options.ssh_server_port,
                            cloud_identifier))
            self._db_conn.commit()
            if cursor.rowcount < 1:
                _log.error('Cannot update Client:')
                await asyncio.sleep(5)
                raise web.HTTPBadRequest
            else:
                _log.debug(f'Updated Client: {cloud_identifier}')
        else:
            _log.debug(f'Created Client: {cloud_identifier}')

        data['ssh_server_hostname'] = self._options.ssh_server_hostname
        data['ssh_server_port'] = self._options.ssh_server_port
        data['ssh_server_listen_port'] = ssh_server_listen_port
        data['cloud_server_url'] = self._options.cloud_server_url

        await self._create_apache_conf()
        await self._create_nginx_conf()
        return web.json_response(data)

    async def _create_apache_conf(self):
        cursor = self._db_conn.cursor()
        cursor.execute("SELECT * FROM clients")
        clients = cursor.fetchall()

        env = Environment(
            loader=FileSystemLoader(
                Path(self._options.apache_tpl_file).parent),
            autoescape=False
        )
        try:
            template = env.get_template(
                Path(self._options.apache_tpl_file).name)
        except TemplateNotFound as e:
            _log.error(f'No Apache template found: {e}')
            return False

        Path(self._options.apache_conf_file).parent.mkdir(
            parents=True, exist_ok=True)

        with open(self._options.apache_conf_file, "w") as f:
            f.write(template.render(clients=clients))

        try:
            proc = await asyncio.create_subprocess_exec(
                '/usr/bin/sudo',
                '/usr/sbin/apachectl', 'graceful',
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE)
            stdout, stderr = await proc.communicate()
            print(f'stdout: {stdout}, stderr: {stderr}')
        except FileNotFoundError as e:
            print(f'Apache not reloaded: {e}')
            return False
        return True

    async def _create_nginx_conf(self):
        cursor = self._db_conn.cursor()
        cursor.execute("SELECT * FROM clients")
        clients = cursor.fetchall()

        env = Environment(
            loader=FileSystemLoader(Path(self._options.nginx_tpl_file).parent),
            autoescape=False
        )
        try:
            template = env.get_template(
                Path(self._options.nginx_tpl_file).name)
        except TemplateNotFound as e:
            _log.error(f'No NGINX template found: {e}')
            return False

        Path(self._options.nginx_conf_file).parent.mkdir(
            parents=True, exist_ok=True)

        with open(self._options.nginx_conf_file, "w") as f:
            f.write(template.render(clients=clients))

        try:
            proc = await asyncio.create_subprocess_exec(
                '/usr/bin/sudo',
                '/bin/systemctl', 'reload', 'nginx',
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE)
            stdout, stderr = await proc.communicate()
            print(f'stdout: {stdout}, stderr: {stderr}')
        except FileNotFoundError as e:
            print(f'NGINX not reloaded: {e}')
            return False
        return True



"""

"""

from yoyo import step

__depends__ = {'__init__'}

steps = [
    step(
    '''CREATE TABLE "users" (
	"id"	        INT GENERATED BY DEFAULT AS IDENTITY,
	"identity"	    VARCHAR,
	"password"	    VARCHAR,
    "full_name"     VARCHAR,
    "email"         VARCHAR,
    "permission_id"	INTEGER,
	"is_superuser"	INTEGER DEFAULT 0,
	"is_disabled"	INTEGER DEFAULT 0,
	"remember_me"	VARCHAR,
	"created_at"	TIMESTAMPTZ DEFAULT CURRENT_TIMESTAMP,
    CONSTRAINT "identity_uc" UNIQUE("identity"),
	PRIMARY KEY("id"))''',
    '''DROP TABLE users'''
    ),
    step(
    '''INSERT INTO "users"
    ("identity","is_superuser")
    VALUES ('admin',1)''',
    """DELETE FROM users WHERE identity = 'admin' """
    ),
    step(
    ''' CREATE TABLE "users_permissions" (
	"id"	        INT GENERATED BY DEFAULT AS IDENTITY,
	"name"	        VARCHAR,
	"user_id"	    INTEGER NOT NULL REFERENCES "users"("id") ON DELETE CASCADE,
	"created_at"	TIMESTAMPTZ DEFAULT CURRENT_TIMESTAMP,
	PRIMARY KEY("id"))''',
    ''' DROP TABLE "users_permissions" '''
    ),
    step(
    ''' CREATE TABLE "users_tokens" (
	"id"	        INT GENERATED BY DEFAULT AS IDENTITY,
	"sequence"	    VARCHAR NOT NULL,
    "hash"	        VARCHAR,
    "category"	    VARCHAR,
	"user_id"	    INTEGER NOT NULL REFERENCES "users"("id") ON DELETE CASCADE,
	"created_at"	TIMESTAMPTZ DEFAULT CURRENT_TIMESTAMP,
    CONSTRAINT "sequence_uc" UNIQUE("sequence"),
	PRIMARY KEY("id"))''',
    ''' DROP TABLE "users_tokens" '''
    ),
    step(
    '''CREATE TABLE "clients" (
	"id"	                    INT GENERATED BY DEFAULT AS IDENTITY,
	"cloud_identifier"	        VARCHAR,
    "is_disabled"               INTEGER DEFAULT 0,
	"public_user_key"    	    VARCHAR,
    "ssh_server_listen_port"    INTEGER,
  	"ssh_server_hostname"       VARCHAR,
    "ssh_server_port"           INTEGER,
	"created_at"	            TIMESTAMPTZ DEFAULT CURRENT_TIMESTAMP,
    CONSTRAINT "cloud_identifier_uc" UNIQUE("cloud_identifier"),
    CONSTRAINT "ssh_server_listen_port_uc" UNIQUE("ssh_server_listen_port"),
	PRIMARY KEY("id"))''',
    '''DROP TABLE clients'''
    )
]


from aiohttp import web
from aiohttp_session import get_session
import aiohttp_jinja2
import bcrypt
import secrets
import asyncio

from logging import getLogger
_log = getLogger(__name__)


async def current_user_ctx_processor(request):
    user = await get_security().get_user(request)
    return {'user': user}


def get_security():
    return Security._instance


class Security():
    _instance = None

    def __init__(self, app, db, options):
        if Security._instance is None:
            Security._instance = self
        self._db_conn = db.conn
        self._options = options
        app.router.add_get('/login', self._login, name='login')
        app.router.add_post('/login', self._login_apply)
        app.router.add_get('/logout', self._logout, name='logout')
        app.router.add_get('/profile', self._profile, name='profile')
        app.router.add_post('/profile', self._profile_apply)

    async def get_user(self, request):
        session = await get_session(request)
        user_id = session.get('user_id')
        if user_id is not None:
            cursor = self._db_conn.cursor()
            cursor.execute("""SELECT *, 'session' AS provider
                            FROM users
                            WHERE id = %s""", (user_id,))
            user = cursor.fetchone()
            # Disabled User must be logged out
            if user is None or user['is_disabled']:
                session.pop('user_id', None)
                return None
            return user
        remember_me = request.cookies.get('remember_me')
        if remember_me is not None:
            cursor.execute("""SELECT *, 'remember_me' AS provider
                            FROM users
                            WHERE remember_me = %s""", (remember_me,))
            user = cursor.fetchone()
            # Disabled User must be logged out
            if user is not None and user['is_disabled']:
                try:
                    cursor.execute("""UPDATE users
                                    SET remember_me = ''
                                    WHERE id = %s""", (user['id'],))
                    self._db_conn.commit()
                except Exception:
                    self._db_conn.rollback()
                return None
            return user
        return None

    async def check_perm(self, request, perm=None):
        user = await self.get_user(request)
        if user is None:
            return None
        if user['is_superuser']:
            return user
        if perm is None:
            return user

        cursor = self._db_conn.cursor()
        cursor.execute('''SELECT * FROM users_permissions
                        WHERE user_id = %s''', (user['id'],))
        rows = cursor.fetchall()
        for row in rows:
            if row['name'] == perm:
                return user
        return None

    async def raise_perm(self, request: web.BaseRequest, perm: str = None):
        """Generate redirection to login form if permission is not
        sufficent. Append query string with information for redirecting
        after login to the original url.

        :param request: [description]
        :type request: web.Baserequest
        :param perm: If no permission is given, check auth only
        :type perm: str, optional
        :raises web.HTTPSeeOther: [description]
        """
        if await self.check_perm(request, perm=perm) is None:
            raise web.HTTPSeeOther(
                request.app.router['login'].url_for(
                ).with_query(
                    {"redirect": str(request.rel_url)}
                )
            )

    @aiohttp_jinja2.template("security/login.html")
    # @router.get("/login", name='login')
    async def _login(self, request: web.Request):
        redirect = request.rel_url.query.get("redirect", "")
        csrf_token = secrets.token_urlsafe()
    #    session = await new_session(request)
        session = await get_session(request)
        session['csrf_token'] = csrf_token
        return {'csrf_token': csrf_token, 'redirect': redirect}

    async def _login_apply(self, request: web.Request):
        # @ router.post("/login")
        form = await request.post()
        session = await get_session(request)
        if session.get('csrf_token') != form['csrf_token']:
            # Normally not possible
            await asyncio.sleep(2)
            raise web.HTTPUnauthorized()

        if form.get('identity') is None:
            await asyncio.sleep(2)
            raise web.HTTPSeeOther(request.app.router['login'].url_for())

        cursor = self._db_conn.cursor()
        cursor.execute("""SELECT *, 'login' AS provider
                        FROM users
                        WHERE identity = %s""", (form['identity'],))
        user = cursor.fetchone()

        if (
            user is None or
            user['is_disabled'] or
            user['password'] is None or
            not len(user['password'])
        ):
            await asyncio.sleep(2)
            raise web.HTTPSeeOther(request.app.router['login'].url_for())
        if not bcrypt.checkpw(form['password'].encode('utf-8'),
                              user['password'].encode('utf-8')):
            await asyncio.sleep(2)
            raise web.HTTPSeeOther(request.app.router['login'].url_for())

        session['user_id'] = user['id']

        location = form.get('redirect')
        if location is None or location == '':
            location = request.app.router['home'].url_for()
        response = web.HTTPSeeOther(location=location)
        if form.get('remember_me') is not None:
            remember_me = secrets.token_urlsafe()
            try:
                cursor.execute('''UPDATE users
                                SET remember_me = %s
                                WHERE id = %s ''', (remember_me, user['id']))
                self._db_conn.commit()
            except Exception as e:
                self._db_conn.rollback()
                _log.error(f'Rememeber_me: unable to update: {e}')
            response.set_cookie("remember_me", remember_me,
                                max_age=self._options.max_age_remember_me,  # noqa: E501
                                httponly=True,
                                samesite='strict')
        return response

    async def _logout(self, request: web.Request):
        # @ router.get("/logout", name='logout')
        user = await self.get_user(request)
        cursor = self._db_conn.cursor()
        if user is not None:
            try:
                cursor.execute("""UPDATE users
                                SET remember_me = ''
                                WHERE id = %s""", (user['id'],))
                self._db_conn.commit()
            except Exception as e:
                _log.error(f'logout: Unable to Update: {e}')
                self._db_conn.rollback()
        session = await get_session(request)
        if session is not None:
            session.pop('user_id', None)
        response = web.HTTPSeeOther(request.app.router['login'].url_for())
    #    response.set_cookie('remember_me', None,
    #                        expires='Thu, 01 Jan 1970 00:00:00 GMT')
        response.del_cookie('remember_me')
        return response

    @ aiohttp_jinja2.template("security/profile.html")
    async def _profile(self, request: web.Request):
        return {}

    async def _profile_apply(self, request: web.Request):
        form = await request.post()
        user = await self.get_user(request)

    # TODO check old_password and equality of pw1 an pw2
        password = form['new_password_1'].encode('utf-8')
        password = bcrypt.hashpw(password, bcrypt.gensalt())
        password = password.decode('utf-8')
        cursor = self._db_conn.cursor()
        try:
            cursor.execute('''UPDATE users
                            SET password = %s
                            WHERE id = %s ''', (password, user['id']))
            self._db_conn.commit()
        except Exception as e:
            _log.error(f'Unable to update: {e}')
            self._db_conn.rollback()
        if cursor.rowcount < 1:
            raise web.HTTPSeeOther(request.app.router['profile'].url_for())
        else:
            raise web.HTTPSeeOther(request.app.router['home'].url_for())
